; hd_io.asm
;
; Interface to BIOS int 0x13 functions
;
; Copyright (C) 2000, Suzhe. See file COPYING for details.
;

%ifndef MAIN
%include "hd_io.h"
%endif

%define HAVE_HD_IO

%define RETRY_TIMES     3

        bits 16
;==============================================================================
;check_int13ext ---- Check if the extension of int13h is presents for a driver.
;input:
;      dl = driver id
;output:
;      cf = 0, int13ext is presents
;          cl = interface support bitmap.
;      cf = 1, int13ext is not presents
;==============================================================================
check_int13ext:
        mov bx, 0x55aa
        mov ah, INT13H_EXT_INSTCHECK
        int 0x13                        ; Check if int13h extension is presents
        jc .not_present
        cmp bx, 0xaa55
        jne .not_present
        test cl, EXT_SUBSET_FIXED       ; Check if this drive supports extended
        jz .not_present                 ; read/write
        clc
        ret
.not_present:
        stc
        ret

;==============================================================================
;get_drive_info ---- Get drive informations
;input:
;      dl = drive id
;      es:di -> buffer for drive info (struc struc_driveinfo)
;output:
;      cf = 0 ok
;      cf = 1 drive is not presents
;==============================================================================
get_drive_info:
        pusha
        
        mov [di + struc_driveinfo.id], dl
        mov byte [di + struc_driveinfo.flags], 0 ; clear the flags

        mov ah, INT13H_GETTYPE
        push dx
        int 0x13
        pop dx
        jc .drive_not_present
        or ah, ah                           ; ah = drive type, = 0 no such drive
        jz .drive_not_present
        
        push es
        push di
        push ax                             ; save drive type
        mov ah, INT13H_GETINFO
        int 0x13
        pop ax
        pop di
        pop es
        jc .drive_not_present

        push ax                             ; drive type.
        push bx                             ; floppy type.
        
        xor ah, ah
        mov al, cl                               ;
        and al, 0x3f                             ; get sectors per track
        mov [di + struc_driveinfo.sectors], ax   ;

        xchg ah, ch                              ;
        shr cx, 6                                ;
        xchg ah, ch                              ;
        ror cx, 8                                ; get cylinders
        inc cx                                   ;
        mov [di + struc_driveinfo.cylinders], cx ;

        xor dl, dl                               ;
        xchg dl, dh                              ; get heads
        inc dx                                   ;
        mov [di + struc_driveinfo.heads], dx     ;

        pop bx
        pop ax

        mov dh, INFOFLAG_DRIVEOK
        cmp ah, DRV_TYPE_FIXED                  ; fixed drive.
        je .fixed_drive
        or dh, INFOFLAG_REMOVABLE
        
.fixed_drive:
        or dh, INFOFLAG_CHSVALID

        mov dl, [di + struc_driveinfo.id]
        push dx
        call check_int13ext
        pop dx
        jc .ext_not_present
        or dh, INFOFLAG_EXTOK
        
;;;;;;;;;;;;;;;;;;; perhaps these will be used in future ;;;;;;;;;;;;;;;;;;;
%if 0
;Int13h is presents, use int13h ext to get drive info
        mov ah, INT13H_EXT_GETINFO
        lea si, [tmp_extparam]
        mov word [si + struc_extparam.pack_size], EXTPARAM_SIZE
        push dx
        int 0x13
        pop dx
        jc .drive_not_present
        
        mov al, [si + struc_extparam.flags]               ;
        and al, (INFOFLAG_CHSVALID + INFOFLAG_REMOVABLE)  ; fill in the
        or al, INFOFLAG_EXTOK + INFOFLAG_DRIVEOK          ; struc_driveinfo.flags
        mov [di + struc_driveinfo.flags], al              ;

        test dh, INFOFLAG_REMOVABLE
        jz .ext_fixed
        xor ax, ax
        jmp .fill_cyls
        
.ext_fixed:
        mov eax, [si + struc_extparam.total_sectors]      ;
        mov edx, [si + struc_extparam.total_sectors + 4]  ; calculate the
        xor ecx, ecx                                      ; number of
        mov cx, [di + struc_driveinfo.sectors]            ; cylinders
        div ecx                                           ;
        xor edx, edx                                      ; use dword to
        mov cx, [di + struc_driveinfo.heads]              ; prevent overflow
        div ecx
                                                          ;
.fill_cyls:
        mov [di + struc_driveinfo.cylinders], ax          ;
%endif
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.ext_not_present:
        mov [di + struc_driveinfo.flags], dh

        popa
        clc
        ret
        
.drive_not_present:
        popa
        stc
        ret

;==============================================================================
;get_drive_flags ---- get drive flags
;input:
;      dl = drive id
;output:
;      cf = 0 ok, al = drive flags
;      cf = 1 error, al = 0
;==============================================================================
get_drive_flags:
        push di
        lea di, [tmp_driveinfo]
        call get_drive_info
        mov al, [di + struc_driveinfo.flags]
        pop di
        ret

;;;;;;;;;;;;;;;;;;; perhaps these will be used in future ;;;;;;;;;;;;;;;;;;;
%if 0
;==============================================================================
;init_drives_info ---- get all drives info
;input:
;      es:di -> buffer to store the info of each drives
;               should have enought space to store 255 drives' info
;output:
;      dh = the number of drives (include the floppy)
;==============================================================================
init_drives_info:
        push es
        push di
        mov cx, 256
        xor dx, dx
.loop_query:
        push cx
        push dx
        call get_drive_info
        pop dx
        pop cx
        jc .bad_drive
        inc dh
        
.bad_drive:
        add di, SIZE_OF_DRIVEINFO
        inc dl
        loop .loop_query
        
        pop di
        pop es
        ret

%endif
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;==============================================================================
;lba2chs ---- lba to chs translation
;input:
;      cx : bx = lba address ( bx is low word )
;      ds : si -> drive info struc
;output:
;      cf = 0 success
;            ch = low eight bits of cylinder number
;            cl = sector number 1-63 (bits 0-5)
;                 high two bits of cylinder (bits 6-7, hard disk only)
;            dh = head number
;      cf = 1 error ( cylinder exceed )
;==============================================================================
lba2chs:
        push ax

        mov ax, cx                  ;
        rol eax, 16                 ; cx : bx -> eax
        mov ax, bx                  ;

        xor ecx, ecx
        mov cx, [si + struc_driveinfo.sectors] ; calculate sector:

        or cx, cx                              ; sectors per track cannot equal
        jz .convert_failed                     ; to zero.
        
        xor edx, edx
        div ecx                                ; sector =  lba % sects_per_track + 1
        inc dx                                 ; lba1 = lba1 / sects_per_track
        mov cx, dx                             ;

                                               ; calculate head and cylinder:
        xor ebx, ebx                           ; head = lba1 % num_of_heads
        mov bl, [si + struc_driveinfo.heads]   ; cylinder = lba1 / num_of_heads

        or bl, bl                              ; heads cannot equal to zero.
        jz .convert_failed                     ;
        
        xor edx, edx
        div ebx                                ;
        
        mov dh, dl                             ; head = dl <= 255
        cmp ax, MAX_CYLS                       ; cylinder = ax
        ja .convert_failed                     ; cylinder is too big.

        xchg al, ah                            ;
        shl al, 6                              ; fill cylinder and sector into
        or cx, ax                              ; cx
        pop ax
        clc
        ret
        
.convert_failed:
        pop ax
        stc
        ret
        
;==============================================================================
;disk_access_low ---- read / write sectors on disk
;input:
;      ah      = function id, ( 02 = read, 03 = write )
;      al      = number of sectors to be accessed
;      bx : cx = lba address ( bx is low word )
;      ds : si -> drive info struc
;      es : di -> buffer to get/store data
;output:
;      cf = 0 success
;      cf = 1 error
;==============================================================================
disk_access_low:
        pusha
        
        mov dh, [si + struc_driveinfo.flags]
        test dh, INFOFLAG_DRIVEOK ; check if drive ok
        jz .access_error
        
        test dh, INFOFLAG_EXTOK
        jz .no_ext
        
;int13h extension is present, use it
        
        mov dl, [si + struc_driveinfo.id]    ; get drive id
        add ah, 0x40                         ; ext read func id = 0x42
        lea si, [tmp_int13ext]
        xor dh, dh
        push dx
        xor edx, edx
        mov byte [si + struc_int13ext.pack_size], 0x10 ;
        mov [si + struc_int13ext.blk_num_high1], edx ; clear and set
        mov [si + struc_int13ext.reserved], dl       ; some stuff.
        mov [si + struc_int13ext.reserved1], dl      ;
        pop dx

.retry_ext_read:
        mov [si + struc_int13ext.blk_count], al
        mov [si + struc_int13ext.buf_addr_off], di
        mov [si + struc_int13ext.buf_addr_seg], es
        mov [si + struc_int13ext.blk_num_low1], bx
        mov [si + struc_int13ext.blk_num_low2], cx
        
        push ax
        push dx
        int 0x13
        pop dx
        pop ax
        jnc .access_ok
        call reset_drive
        inc dh
        cmp dh, RETRY_TIMES                 ; retry 3 times
        jb .retry_ext_read
        jmp .access_error

;int13h extension is not ok, use old int13h
.no_ext:
        call lba2chs
        jc .access_error
        mov dl, [si + struc_driveinfo.id]   ; dl = drive id
        mov bx, di                          ; es : bx -> buffer
        xor di, di
        
.retry_read:
        push ax
        push dx
        int 0x13
        pop dx
        pop ax
        jnc .access_ok
        call reset_drive
        inc di
        cmp di, RETRY_TIMES                 ; retry 3 times
        jb .retry_read
        jmp .access_error

.access_error:
        stc
.access_ok:
        popa
        ret

;==============================================================================
;disk_access ---- read / write sectors on disk
;input:
;      ah = function id, ( 02 = read, 03 = write )
;      al = number of sectors to be accessed
;      cx : bx = lba address ( cx = high word, bx = low word )
;      dl = drive id
;      es : di -> buffer
;output:
;      cf = 0 success
;      cf = 1 error
;==============================================================================
disk_access:
        push si
        push di

        push es
        push ds                         ; ds -> es, tmp_driveinfo is in ds segment.
        pop es
        
        lea di, [tmp_driveinfo]
        call get_drive_info
        
        pop es
        pop di
        
        jc .access_error
        lea si, [tmp_driveinfo]
        call disk_access_low
.access_error:
        pop si
        ret

;==============================================================================
;reset_drive ---- reset the drive
;input:
;      dl = drive id
;output:
;      cf = 0 success
;      cf = 1 error
;==============================================================================
reset_drive:
        pusha
        xor ax, ax
        int 0x13
        popa
        ret

%ifndef MAIN
%include "tempdata.asm"
%endif

